package enos.scrabble.domain;

import java.util.Arrays;
import enos.scrabble.util.BigramEvaluator;

public class Word {
  private final String theWord;
  private String theWordWithCaps = null;
  private Anchor theAnchor;
  private float bigramScore = -1;
  private int totalLetterScores = -1;
  private int hashCode = 0;
  private byte[] encoding = new byte[27];
  private boolean usesSeven = false;
  private String lettersUsed = "";
  /**
   * Our constructor uses the string.intern() method to assure that the string
   * we're creating is from the pool of Strings stored on the heap. This makes
   * for very fast comparisons, since we can now use == rather than .equals()
   * for String comparisons, since any two identical strings now refer to the
   * same object in the symbol table.
   */
  public Word(String aWord) {
    theWord = (aWord.toLowerCase()).intern();
    // take into account words with capital letters, representing the blank tile
    if (!theWord.equals(aWord)) theWordWithCaps = aWord;
    encode();
  }

  public Word(Word aWord, Anchor anAnchor) {
    this(aWord.theWord + anAnchor.getLetters());
    this.theWordWithCaps = aWord.theWordWithCaps;
    this.theAnchor = anAnchor;
  }

  public void increaseBigramScore(float increase) {
    if (-1 == bigramScore) bigramScore = 0;
    bigramScore += increase;
  }

  public void addAnchor(Anchor an) {
    theAnchor = an;
  }

  public float getBigramScore() {
    if (-1 == bigramScore) {
      BigramEvaluator.singleton().scoreBigrams(this);
    }
    return bigramScore;
  }

  public int hashBinary() {
    // lazy init, only do this once
    if (0 == hashCode) {
      int length = theWord.length();
      boolean[] countedChars = new boolean[26];
      for (int i = 0; i < length; i++) {
        char theChar = theWord.charAt(i);
        int charValue = theChar - 'a';
        // we only count one occurrence of each char, so that
        // "good" and "god" would hash to the same bucket
        if (!countedChars[charValue]) {
          hashCode += (int) Math.pow(2, charValue);
          countedChars[charValue] = true;
        }
      }
    }
    return hashCode;
  }

  public boolean containsSubstring(String sub) {
    return (theWord.indexOf(sub.toLowerCase()) > -1);
  }

  public boolean containsBlank() {
    return (theWord.indexOf("{") > -1);
  }

  public boolean hasAnchor() {
    return (null != theAnchor);
  }

  public Anchor getAnchor() {
    return theAnchor;
  }

  public String getAnchorLetters() {
    return theAnchor.getLetters();
  }

  public String anchorString() {
    return theAnchor.getLetters();
  }

  public int anchorLength() {
    if (null != theAnchor)
      return theAnchor.getLetters().length();
    else return -1;
  }

  public int indexOfAnchor() {
    if (hasAnchor())
      return this.toString().indexOf(theAnchor.getLetters());
    else
      return -1;
  }

  public boolean usesSeven() {
    if (this.hasAnchor())
      return ( (theWord.length() - theAnchor.length()) == 7);
    else
      return (theWord.length() == 7);
  }

  public void setLettersUsed(String letters) {
    lettersUsed = letters;
  }

  public String newLettersUsed() {
    return lettersUsed;
/*    if (this.hasAnchor()) {
      StringBuffer buf = new StringBuffer(this.toString());
      char[] letts = theAnchor.getLetters().toCharArray();
      for (int i = 0; i < letts.length; i++) {
        int index = -1;
        if (-1 < (index = buf.indexOf((letts[i] + "")))) {
          buf.deleteCharAt(index);
        }
      }
      return buf.toString();
    }
    else {
      return this.toString();
    }
  */
  }

  public int totalLetterScores() {
    if ( -1 == totalLetterScores) {
      totalLetterScores = 0;
      //this is for counting the blank as 0
//      char[] letters = (null != theWordWithCaps) ? theWordWithCaps.toCharArray()
//          : theWord.toCharArray();
      char[] letters = theWord.toCharArray();
      for (int i = 0; i < letters.length; i++) {
        totalLetterScores += Game.letterScore(letters[i]);
      }
      if ( (null == theAnchor && theWord.length() == 7)
           || (null != theAnchor && theWord.length() - theAnchor.length() == 7)) {
        totalLetterScores += 50;
      }
    }
    return totalLetterScores;
  }

  public double getWordPotential() {
//    return getBigramScore() + totalLetterScores();
//    return getBigramScore() * totalLetterScores();
//    return totalLetterScores();
      return Math.exp(getBigramScore()) * totalLetterScores();
//    return getBigramScore();
  }

  public String lettersWithoutBlank() {
    if (encoding[26] < 1) {
      return theWord;
    } else {
      return theWord.replaceAll("\\{", "");
    }
  }

  public int length() {
    return theWord.length();
  }

  private void encode() {
    int length = theWord.length();
    for (int i = 0; i < length; i++) {
      char theChar = theWord.charAt(i);
      encoding[theChar - 'a']++;
    }
  }

  public boolean equals(Object other) {
    if (other == this) {
      return true;
    }
    if (!(other instanceof Word)) {
      return false;
    }
    Word that = (Word)other;
    // see constructor comment for why this works
    if (this.theWord == that.theWord) {
      return true;
    }else {
      return false;
    }
  }

  public boolean sameLetters(Word that) {
    if (Arrays.equals(this.encoding, that.encoding)) {
      return true;
    }else {
      return false;
    }
  }

  public int hashCode() {
    return theWord.hashCode();
  }

  public boolean wordEquals(Word that) {
    // see constructor comment for why this works
    if (this.theWord == that.theWord) {
      return true;
    }else {
      return false;
    }
  }

  public void synchronizeCaps(Word that) {
    if (null != that.theWordWithCaps) {
      char[] target = this.theWord.toCharArray();
      char[] source = that.theWordWithCaps.toCharArray();
      for (int i = 0; i < source.length; i++) {
        if (source[i] <= 'Z') {
          int index = this.theWord.indexOf((source[i] + "").toLowerCase());
          target[index] = source[i];
        }
      }
      this.theWordWithCaps = (new String(target)).intern();
    }
  }

  public String toString() {
    if (null == theWordWithCaps)
      return theWord;
    else
      return theWordWithCaps;
  }
}